local L = RC_locale()
local addonName = "RetardCheck"
local LaddonName = L[addonName]
RetardCheck = LibStub("AceAddon-3.0"):NewAddon(addonName)
local addon = RetardCheck 
addon.svnrev = {}
addon.svnrev["RetardCheck.lua"] = tonumber(("$Revision: 29 $"):match("%d+"))

local defaults = {
  profile = {
    general = {
      debug = false, -- for addon debugging
      sound = true,
      soundrandom = true,
      soundname = "Troxed 1",
      chatalert = true,
      alwayschat = false,
      readycheck = true,
      targetboss = true,
      autochecktime = 5,
    },
    minimap = {
        hide = false,
    },
    checks = {
      gear = true,
      stat = true,
      buff = true,

      buffduration = 5,
      
      dkhorn = false,
      shamanshield = false, 
      warriorshout = false,
    },
    thresh = {
      ilvl = 250,
      durability = 0.10,
      shamanimbue = 5,
      roguepoison = 5,
      resilience = 300,
      spellhit = 0.005,
      meleehit = 0.005,
      rangedhit = 0.005,
      expertise = 0.005,
    },
  }
}

local settings = defaults.profile
local optionsFrame, profileFrame
local check_class
addon.scantt = CreateFrame("GameTooltip", addonName.."_ScanTooltip", UIParent, "GameTooltipTemplate")
addon.scantt:SetOwner(UIParent, "ANCHOR_NONE");
local revision = tonumber(("$Revision: 29 $"):match("%d+"))
local minimapIcon = LibStub("LibDBIcon-1.0")
local LDB

local function chatMsg(msg)
     DEFAULT_CHAT_FRAME:AddMessage("\124cff9999ff"..addonName.."\124r: "..msg)
end
local function chatAlert(msg)
   if settings.general.chatalert then
     chatMsg(msg)
   end
end
local function debug(msg)
  if addon.db.profile.general.debug then
     chatMsg(msg)
  end
end

addon.soundfiles = {
  { "Ermey", 9 },
  { "Faelagund", 5 },
  { "Onyxia", 2 },
  { "Letomi", 3 },
  { "Sinnerman", 1 },
  { "Carrey", 1 },
  { "Troxed", 11 },
  "Sound/creature/Illidan/BLACK_Illidan_04.ogg",
  "Sound/Creature/XT002Deconstructor/UR_XT002_Special01.wav",
  "Sound/Creature/Moroes/MoroesSlay03.ogg",
}
addon.soundtable = {} -- name=>filename and index=>filename
addon.soundok = "Interface\\AddOns\\"..addonName.."\\sound\\tada.mp3"
for _,v in pairs(addon.soundfiles) do
 if type(v) == "table" then
   local name = v[1]
   for i = 1,v[2] do
     local filename = "Interface\\AddOns\\"..addonName.."\\sound\\"..name:gsub(" ","_").."_"..i..".mp3"
     addon.soundtable[name.." "..i] = filename
     table.insert(addon.soundtable, filename)
   end
 else
   local name = v:gsub("^.+/([^/]+)$","%1"):gsub("%....$",""):gsub("_"," ")
   local filename = v
   addon.soundtable[name] = filename
   table.insert(addon.soundtable, filename)
 end
end

function addon:myOptions() 
return {
  type = "group",
  set = function(info,val)
          local s = settings ; for i = 1,#info-1 do s = s[info[i]] end
          s[info[#info]] = val; debug(info[#info].." set to: "..tostring(val))
          addon:Update()
        end,
  get = function(info)
          local s = settings ; for i = 1,#info-1 do s = s[info[i]] end
          return s[info[#info]] end,
  args = {
   general = {
    type = "group",
    inline = true,
    name = L["General"],
    args = {
      debug = {
        name = L["Debug"],
        desc = L["Toggle debugging output"],
        type = "toggle",
        guiHidden = true,
      },
      config = {
        name = L["Config"],
        desc = L["Open the configuration GUI"],
        type = "execute",
        guiHidden = true,
        func = function() addon:Config() end,
      },
      check = {
        name = L["Check"],
        desc = L["Check now to see if you are retarded"],
        type = "execute",
        guiHidden = true,
        func = function() addon:RunCheck(true) end,
      },
      minimap = {
        order = 15,
        name = L["Minimap Icon"],
        desc = L["Display minimap icon"],
        type = "toggle",
        set = function(info,val)
          settings.minimap.hide = not val
          addon:Update()
	end,
        get = function() return not settings.minimap.hide end,
      },
      checkheader = {
        order = 16,
	name = L["Automated Checks"],
	type = "header",
      },
      readycheck = {
        order = 17,
        name = L["Ready Check"],
        desc = L["Check if you're retarded during ready checks"],
        type = "toggle",
      },
      targetboss = {
        order = 18,
        name = L["Boss Target"],
        desc = L["Check if you're retarded when you first target a boss"],
        type = "toggle",
      },
      autochecktime = {
        order = 19,
        name = L["Auto-check interval"],
        desc = L["Minimum number of minutes between automatic checks"],
        type = "range",
        step = 0.5,
        min = 0,
        max = 60,
      },
      alertheader = {
        order = 20,
	name = L["Alerts"],
	type = "header",
      },
      sound = {
        order = 25,
        name = L["Alert Sound"],
        desc = L["Play an alert sound when a retard is detected"],
        type = "toggle",
      },
      soundrandom = {
        order = 27,
        name = L["Random Sound"],
        desc = L["Use a randomly-chosen sound"],
	disabled = function() return not settings.general.sound end,
        type = "toggle",
      },
      soundname = {
        order = 29,
        name = L["Sound"],
        desc = L["Sound to play when a retard is detected"],
	disabled = function() return not settings.general.sound or settings.general.soundrandom end,
        type = "select",
	style = "dropdown",
	values = function() 
	  local t = {}
	  for n,_ in pairs(addon.soundtable) do
	    if not tonumber(n) then
	      t[n] = n
	    end
	  end
	  return t
	end,
	set = function(info,val) 
	  settings.general.soundname = val
	  if val ~= random then
	    PlaySoundFile(addon.soundtable[val])
	  end
	end,
      },
      chatalert = {
        order = 30,
        name = L["Chat Alert"],
        desc = L["Print an alert in chat when a retard is detected"],
        type = "toggle",
      },
      alwayschat = {
        order = 35,
        name = L["Always Chat"],
        desc = L["Print to chat whenever a retard check is performed"],
        type = "toggle",
      },
    },
   },
   gearchecks = {
	name = L["Gear Checks"],
	type = "group",
	order = 1,
	args = addon:getCheckOptions("gear", L["Gear Checks"]),
   },
   statchecks = {
	name = L["Stat Checks"],
	type = "group",
	order = 2,
	args = addon:getCheckOptions("stat", L["Stat Checks"]),
   },
   buffchecks = {
	name = L["Buff Checks"],
	type = "group",
	order = 3,
	args = addon:getCheckOptions("buff", L["Buff Checks"]),
   },
  }
} 
end

function addon:getCheckOptions(checktype, groupname)
  local ret = {}
  ret[checktype] = {
      order = 0,
      name = groupname,
      type = "toggle",
      width = (checktype == "buff" and "double") or "full",
  }
  -- special cases
  if checktype == "buff" then
    ret.buffduration = {
      order = 0.1,
      name = L["Buff Duration"],
      desc = L["Minimum time remaining in minutes to consider the buff valid"],
      type = "range",
      min = 0,
      max = 15,
      step = 1,
      disabled = function() return not settings.checks[checktype] end,
    }
  end
  for o,v in pairs(addon.checks_ordered) do
   local ct = v.checktype or "gear"
   if (ct == checktype) then
    local desc = v.desc or v.failname
    --if v.playeronly then desc = desc.." "..L["(player only)"]
    if v.header then
    ret[v.checkname] = {
      order = o,
      name = v.header,
      type = "header",
    }
    else
    ret[v.checkname] = {
      order = o,
      name = v.failname,
      desc = desc,
      type = "toggle",
      disabled = function() return not settings.checks[checktype] end,
    }
    if v.threshdesc then
     if ct ~= "buff" then
      ret[v.checkname].order = o*10 + 10002
      ret[v.checkname.."_header"] = {
        order = o*10 + 10001,
	type = "header",
	name = "",
	cmdHidden = true,
      }
     end
      ret[v.checkname.."_thresh"] = {
        order = ret[v.checkname].order+0.5,
	name = v.threshname or v.failname.." "..L["threshold"],
	desc = v.threshdesc,
	step = v.threshstep,
	isPercent = v.threshpercent,
	disabled = function() return not settings.checks[checktype] or not settings.checks[v.checkname] end,
        type = "range",
	min = v.threshmin,
	max = v.threshmax,
        get = function() return settings.thresh[v.checkname] end,
        set = function(info,val)
	        settings.thresh[v.checkname] = val
                addon:Update()
              end
      }
    end -- thesh
    end -- header
   end -- checktype
  end -- for
  return ret
end

local function table_clone(t)
  if not t then return nil
  elseif type(t) == "table" then
    local res = {}
    for k,v in pairs(t) do
      res[table_clone(k)] = table_clone(v)
    end
    return res
  else
    return t
  end
end

local function order_table(t) -- sort and compact a table by order field
    local tmp = {}
    local index = {}
    for k,v in pairs(t) do
       v.checkname = k 
       if v.order then
         if tmp[v.order] then
           debug("ERROR: duplicate table order value: "..v.order)
         else
           tmp[v.order] = v
         end
	 table.insert(index, v.order)
       end
    end
    local max = table.maxn(tmp)
    for k,v in pairs(t) do
       if not v.order then
          max = max + 1
          table.insert(retval, max, v)
          v.order = max
	  table.insert(index, v.order)
       end
    end
    table.sort(index)
    local retval = {}
    for i,o in ipairs(index) do
      retval[i] = tmp[o]
      retval[i].order = i
    end
    return retval
end

-- -------------------------------------------------------------------------------------------------
-- Item queries and helper functions
local FPsubclass
function addon:isFishingPole(itemid)
  if not FPsubclass then
    FPsubclass = select(7, GetItemInfo(6256))
  end
  return select(7, GetItemInfo(itemid)) == FPsubclass
end

local OHsubclass
function addon:isHeldInOffhand(itemid)
  if not OHsubclass then
    OHsubclass = select(7, GetItemInfo(43656))
  end
  return select(7, GetItemInfo(itemid)) == OHsubclass
end
   
local TwoHandsubclass = nil
function addon:is2HWeapon(itemid)
  if not itemid then return false end
  if not TwoHandsubclass then
      TwoHandsubclass = {}
      for _,id in pairs({2491, 2489, 2495, 2493, 6367, 44654}) do
        local subclass = select(7, GetItemInfo(id))
        if subclass and TwoHandsubclass then -- might fail on first call due to lag
          TwoHandsubclass[subclass] = true
        else
          TwoHandsubclass = nil
        end
      end
  end
  local subclass = select(7, GetItemInfo(itemid))
  if TwoHandsubclass then
     return TwoHandsubclass[subclass]
  else
     return true -- conservative don't know
  end
end

local armorSubClass = {
  CLOTH = 3602,
  LEATHER = 2122,
  MAIL = 2387,
  PLATE = 62254,
  first = true,
}
local classArmorType = {
  MAGE = "CLOTH",
  WARLOCK = "CLOTH",
  PRIEST = "CLOTH",
  DRUID = "LEATHER",
  ROGUE = "LEATHER",
  SHAMAN = "MAIL",
  HUNTER = "MAIL",
  PALADIN = "PLATE",
  WARRIOR = "PLATE",
  DEATHKNIGHT = "PLATE",
}
function addon:armorType(itemid)
  if armorSubClass.first then
    local init = true
    for t,n in pairs(armorSubClass) do
      if type(n) == "number" then
        local subclass = select(7, GetItemInfo(n))
	if subclass then 
	  armorSubClass[subclass] = t
	else
	  init = false
	end
      end
    end
    if init then armorSubClass.first = nil end
  end
  local subclass = select(7, GetItemInfo(itemid))
  return armorSubClass[subclass]
end

local buffids = {
  fort = {
        469, -- Commanding shout (must be first)
        21562, -- Prayer of Fortitude
        6307, -- Blood Pact
  },
  kings = {
        20217, -- Blessing of Kings (must be first)
        1126, -- Mark of the Wild
        90363, -- Embrace of the Shale Spider
  },
  might = {
        19740, -- Blessing of Might (must be first)
  },
  strength = {
   	6673, -- Battle Shout (must be first)
   	57330, -- Horn of Winter
        8076, --  Strength of Earth
        93435, -- Roar of Courage
  },
}
local function hasBuff(unitid, buffs)
  for _,id in pairs(buffs) do
    local n = GetSpellInfo(id)
    local name, _,_,_,_,duration,expires, caster, _,_, spellid = UnitBuff(unitid, n)
    -- if spellid == id then  -- too restrictive
    if name == n then -- allow aliases
       if duration and duration > 5*60 then -- check time on buffs over 5 min long
         if expires - GetTime() < settings.checks.buffduration*60 then
	   addon.lowduration = true
	   return false
	 end
       end
       return caster 
    end
  end
  return nil
end
local function buffHeader(Order, Name)
 return {
    order = Order,
    checktype = "buff",
    header = Name,
 }
end
local function buffCheck(Order, Failname, Class, buffs)
 return {
    order = Order,
    failname = Failname,
    checktype = "buff",
    class = Class,
    aliveonly = true,
    playercheck = function(unitid)
      return hasBuff(unitid, buffs)
    end,
  }
end
local function buffCheckTwo(Order, Failname, Class, buff1, buff2)
 return {
    order = Order,
    failname = Failname,
    checktype = "buff",
    class = Class,
    aliveonly = true,
    playercheck = function(unitid)
      local c1 = hasBuff(unitid, buff1)
      local c2 = hasBuff(unitid, buff2)
      if c1 and c2 then -- have both
        return true
      elseif UnitIsUnit(c1 or "none", unitid) or UnitIsUnit(c2 or "none", unitid) then -- already buffed one
        return true
      elseif not c1 then
        return false, tostring(GetSpellInfo(buff1[1]))
      elseif not c2 then
        return false, tostring(GetSpellInfo(buff2[1]))
      end 
    end,
  }
end

local PID_BS   = 2018
local PID_JC   = 25229
local PID_ENG  = 4036
local PID_ENCH = 7411
local PID_INSC = 45357
local PID_LW   = 2108
local PID_TAIL = 3908

local profenchants = {
  [4115]=1, [4116]=1, [4118]=1, -- Tailoring
  [4078]=1, [4079]=1, [4080]=1, [4081]=1, -- Enchanting
  [4189]=1, [4190]=1, [4191]=1, [4192]=1, -- Leatherworking
  [4193]=1, [4194]=1, [4195]=1, [4196]=1, -- Inscription
  [4179]=1, [4180]=1, [4181]=1, [4182]=1, [4183]=1, -- Engineering Gloves
  [4187]=1, [4188]=1, -- Engineering Belt
}

-- returns rank
function addon:hasProfession(pid)
  local pname = GetSpellInfo(pid)
  for _,profid in pairs({ GetProfessions() }) do
    local pn, _, rank, rankmax = GetProfessionInfo(profid)
    if (pn == pname) then return rank end
  end
  return 0
end

local function checkWeaponImbue(unitid,thresh)
      local MH = GetInventoryItemLink(unitid, INVSLOT_MAINHAND)
      local OH = GetInventoryItemLink(unitid, INVSLOT_OFFHAND)
      local hasMainHandEnchant, mainHandExpiration, _,
            hasOffHandEnchant, offHandExpiration, _ = GetWeaponEnchantInfo()
      if MH and (not hasMainHandEnchant or mainHandExpiration < settings.thresh[thresh]*60000) then
        return false, L["Missing weapon imbue on"].." "..MH.." ("..addon.itemslots.MainHandSlot.shortname..")"
      end
      if (check_class == "ROGUE" or GetPrimaryTalentTree(false, false) == 2) and
         OH and (not hasOffHandEnchant or offHandExpiration < settings.thresh[thresh]*60000) then
        return false, L["Missing weapon imbue on"].." "..OH.." ("..addon.itemslots.SecondaryHandSlot.shortname..")"
      end
      return true
end

-- Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging, Name
function addon:decomposeItemLink(itemlink)
  return string.match(itemlink, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
end

function addon:scanBegin(itemlink)
  addon.scantt:ClearLines()
  addon.scantt:SetOwner(UIParent, "ANCHOR_NONE");
  addon.scantt:SetHyperlink(itemlink)
  return addon.scantt:NumLines()
end


function addon:scanLine(i)
  return getglobal(addon.scantt:GetName() .. "TextLeft"..i)
end
function addon:scanText(i)
  local line = addon:scanLine(i)
  local text = line and line:GetText()
  text = text or ""
  return text
end

-- returns true if the item contains an inactive meta
local Lgem = string.match(ENCHANT_CONDITION_EQUAL_VALUE,"\1244([^:]+):[^;]+;")
local Lgems = string.match(ENCHANT_CONDITION_EQUAL_VALUE,"\1244[^:]+:([^;]+);")
function addon:metaInactive(itemlink)
  for i=1,addon:scanBegin(itemlink) do
    local text = addon:scanText(i)
    if text:match(ENCHANT_CONDITION_REQUIRES) and 
       (text:match(Lgem) or text:match(Lgems)) and 
       text:match("\124cff808080") then
        return true
    end
  end
  return false
end  


-- returns number of empty sockets of each color in the unmodified item
local sockettext = {
  EMPTY_SOCKET_RED, EMPTY_SOCKET_YELLOW, EMPTY_SOCKET_BLUE, EMPTY_SOCKET_PRISMATIC,
  EMPTY_SOCKET_META, EMPTY_SOCKET_COGWHEEL, EMPTY_SOCKET_HYDRAULIC,
}
local socketcount = {}
function addon:countEmptySockets(itemlink)
  for i=1,#sockettext do socketcount[i] = 0 end
  local total=0
 
  for i=1,addon:scanBegin(itemlink) do
    local text = addon:scanText(i)
    for i=1,#sockettext do 
        if text:match(sockettext[i]) then
           socketcount[i] = socketcount[i] + 1
           total = total + 1
           break
        end
    end
  end 

  return total, unpack(socketcount)
end

-- return the number of sockets in the raw (ungemmed/unmodded) version of this item
function addon:countRawSockets(itemlink)
  local strippedlink = itemlink:gsub("(item:%d+:%d*):%d*:%d*:%d*:%d*:","%1:0:0:0:0:")
  return addon:countEmptySockets(strippedlink)
end

-- return the number of socketed gems in an item
local Lchimera
local maxgems = 4
function addon:countGems(itemlink)
  if not Lchimera then
    Lchimera = GetItemInfo(52196)
  end
  local tot,chim = 0,0
  for i=1,maxgems do
    local gname,glink = GetItemGem(itemlink, i)
    if gname then
      tot = tot + 1
    end
    if gname and Lchimera and gname:match(Lchimera) then
      chim = chim + 1
    end
  end
  return tot, chim 
end

-- returns raid role for the unit (TANK/HEALER/DAMAGER)
function addon:raidRole(unitid)
  local currrole = UnitGroupRolesAssigned(unitid)
  if currrole and currrole ~= "NONE" then
    return currrole
  end
  if GetPartyAssignment("MAINTANK", unitid) then
    return "TANK"
  end
  local inspect = not UnitIsUnit(unitid,"player")
  local tabIndex = GetPrimaryTalentTree(inspect, false)
  if not tabIndex then return end -- untalented
  local role1,role2 = GetTalentTreeRoles(tabIndex,inspect,false)
  if not role2 then
    return role1
  else -- more than one possibility (eg feral druid)
    local _, class = UnitClass(unitid)
    if class == "DRUID" then
      local tanktalents = 0 -- look for tank talents
      for ti = 1, GetNumTalents(tabIndex, inspect, nil) do
        local link = GetTalentLink(tabIndex, ti, false, nil, nil)
        if link:match("\124Htalent:8293:2\124") or link:match("\124Htalent:8758:1\124") then
          tanktalents = tanktalents + 1
        end
      end
      if tanktalents >= 2 then
        return "TANK"
      else
        return "DAMAGER"
      end
    end
    return nil
  end
end

local outdoor_bg
function addon:inBG()
  if not outdoor_bg then
    outdoor_bg = {}
    for i = 1,2 do
       local _,n = GetWorldPVPAreaInfo(i)
       outdoor_bg[n] = i
    end
  end
  local i, t = IsInInstance()

  if t == "none" then
    local id = outdoor_bg[GetRealZoneText()]
    if id then
      local _,_, isActive = GetWorldPVPAreaInfo(id)
      if isActive then
        return true
      end
    end
  elseif t == "pvp" then
    return true
  else
    return false
  end
end

local dpsstats = {
    MAGE     = { spellhit=true },
    WARLOCK  = { spellhit=true },
    PRIEST   = { spellhit=true },
    HUNTER   = { rangedhit=true },
    DEATHKNIGHT = { meleehit=true, spellhit=true, expertise=true },
    ROGUE    = { meleehit=true, spellhit=true, expertise=true },
    PALADIN  = { meleehit=true, expertise=true },
    WARRIOR  = { meleehit=true, expertise=true },
    SHAMAN   = { [1] = { spellhit=true }, [2] = { meleehit=true, spellhit=true, expertise=true } }, 
    DRUID    = { [1] = { spellhit=true }, [2] = { meleehit=true, expertise=true } }, 
}

function addon:dpsStat(unitid, statname)
    local _, class = UnitClass(unitid)
    local inspect = not UnitIsUnit(unitid,"player")
    local tabIndex = GetPrimaryTalentTree(inspect, false)
    local stats = dpsstats[class][tabIndex] or dpsstats[class]
    return stats[statname]
end

function addon:InitData()
  -- init
  if not addon.itemchecks_ordered then -- first call
    for name,data in pairs(addon.itemslots) do
      data.id = GetInventorySlotInfo(name)
      data.shortname = name:gsub("Slot",""):gsub("[01]",""):gsub("Secondary","Off")
    end
    addon:InitChecks()
    addon.checks_ordered = order_table(addon.checks)
    addon.checktmp = {}
  end

  addon:isFishingPole(6256)   
  addon:is2HWeapon(6256)   
  addon:isHeldInOffhand(6256) 
  addon:armorType(3602)
  addon:countGems("")
  addon:inBG()
end
-- -------------------------------------------------------------------------------------------------
-- slot table
addon.itemslots = {
	ChestSlot    = { enchantable=true }, -- INVSLOT_CHEST
	FeetSlot     = { enchantable=true }, -- INVSLOT_FEET
	WristSlot    = { enchantable=true }, -- INVSLOT_WRIST
	HandsSlot    = { enchantable=true }, -- INVSLOT_HAND
	HeadSlot     = { enchantable=true }, -- INVSLOT_HEAD
	LegsSlot     = { enchantable=true }, -- INVSLOT_LEGS
	ShoulderSlot = { enchantable=true }, -- INVSLOT_SHOULDER
	BackSlot     = { enchantable=true }, -- INVSLOT_BACK
	MainHandSlot = { enchantable=true }, -- INVSLOT_MAINHAND
	SecondaryHandSlot = { enchantable=true }, -- INVSLOT_OFFHAND
	RangedSlot   = { enchantable=true }, -- INVSLOT_RANGED
	Finger0Slot  = {  }, -- INVSLOT_FINGER1
	Finger1Slot  = {  }, -- INVSLOT_FINGER2
	WaistSlot    = {  }, -- INVSLOT_WAIST
	NeckSlot     = {  }, -- INVSLOT_NECK
	Trinket0Slot = {  }, -- INVSLOT_TRINKET1
	Trinket1Slot = {  }, -- INVSLOT_TRINKET2
}
function addon:enchantableSlots() 
  local t = {} 
  for _,s in pairs(addon.itemslots) do
    if s.enchantable then
      table.insert(t, s.shortname)
    end
  end
  return t
end
-- -------------------------------------------------------------------------------------------------
-- Item Checks
local function statCheck(Order, Statname, Failchance, Failname, Threshdesc)
 return {
    order = Order,
    failname = Failname,
    playeronly = true,
    checktype = "stat",
    threshdesc = Threshdesc,
    threshmin = 0,
    threshmax = 0.20,
    threshstep = 0.0025,
    threshpercent = true,
    threshname = L["Miss allowance"],
    playercheck = function(unitid) 
      if addon:raidRole(unitid) ~= "DAMAGER" or -- not a dps
         not addon:dpsStat(unitid, Statname) then -- not a relevant stat
	 return true
      end
      local max = 100*settings.thresh[Statname]
      local lvldelta
      if UnitInRaid(unitid) then
        lvldelta = 3
      else
        lvldelta = 2
      end
      local tgtlvl = UnitLevel(unitid) + lvldelta
      local a,b = Failchance(lvldelta, true)
      if b and b > a and GetInventoryItemID(unitid, INVSLOT_OFFHAND) then -- offhand dodge
         a = b 
      end 
      if a > max then
        return false, string.format("%.02f%% "..MISS.." ("..string.format(L["Level %d target"],tgtlvl)..")", a)
      else
        return true
      end
    end,
  }
end

function addon:InitChecks() 
addon.checks = {

  empty = {
    order = 10,
    failname = L["Empty Slot"],
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
      return itemid and itemlink
    end,
  },

  fishingpole = {
    order = 20,
    failname = L["Fishing Pole"],
    slots = { "MainHand" },
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
        return not addon:isFishingPole(itemid) 
    end,
  },

  ilvl = {
    order = 30,
    failname = L["Item Ilvl"],
    threshdesc = L["Minimum equipped item level to allow"],
    threshmin = 25,
    threshmax = 400,
    threshstep = 1,
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
      local min = settings.thresh.ilvl
      local quality, ilvl = select(3, GetItemInfo(itemid))
      return (ilvl and ilvl >= min) or (quality == 7) -- allow heirlooms
      --- XXX: tooltip scan for ITEM_LEVEL_RANGE_CURRENT to determine level appropriate
    end,
  },

  armortype = {
    order = 35,
    failname = L["Armor Type"],
    desc = L["Check that armor type is class-appropriate for armor specialization"],
    itemstop = true, -- stop checking this slot on failure
    condition = function(unitid)
      return UnitLevel(unitid) >= 40
    end,
    itemcheck = function(itemid, itemlink, slotid)
      local atype = addon:armorType(itemid)
      if atype and atype ~= classArmorType[check_class] then
        if slotid == INVSLOT_BACK then 
	   return true
	else
          return false
	end
      else
        return true
      end
    end
  },

  durability = {
    order = 40,
    playeronly = true,
    failname = L["Item Durability"],
    threshdesc = L["Minimum item durability percentage to allow"],
    threshmin = 0,
    threshmax = 1,
    threshpercent = true,
    itemcheck = function(itemid, itemlink, slotid)
      local cur, max = GetInventoryItemDurability(slotid)
      if not cur then return true end -- slot with no repair or info not available
      local val = cur/max
      local min = settings.thresh.durability
      return val >= min
    end,
  },

  enchant = {
    order = 50,
    failname = L["Item Enchant"],
    slots = addon:enchantableSlots(),
    itemcheck = function(itemid, itemlink, slotid, unitid)
      local enchant = select(4,addon:decomposeItemLink(itemlink))
      local _,_,_,ilvl = GetItemInfo(itemid)
      local _, class = UnitClass(unitid)
      if enchant and tonumber(enchant) and tonumber(enchant) > 0 then
        return true -- XXX: check for "good" enchants
      elseif slotid == INVSLOT_OFFHAND then
        return not addon:isHeldInOffhand(itemid)
      elseif slotid == INVSLOT_RANGED and class ~= "HUNTER" then
        return true 
      elseif slotid == INVSLOT_SHOULDER and 
         not (UnitLevel(unitid) >= 60 and ilvl >= 60) and
	 not (UnitLevel(unitid) >= 64) then 
         -- lvl 60 player/item heavy knothide kit, lvl 64 for BC scryer/aldor inscriptions
        return true
      elseif slotid == INVSLOT_HEAD and 
         not (UnitLevel(unitid) >= 60 and ilvl >= 60) and
	 not (UnitLevel(unitid) >= 70) then 
         -- lvl 60 player/item heavy knothide kit, lvl 70 for BC rep arcanums
        return true
      else
        return false
      end
    end,
  },

  enchantreq = {
    order = 55,
    failname = L["Requirements Not Met"],
    playeronly = true,
    slots = addon:enchantableSlots(),
    itemcheck = function(itemid, itemlink, slotid, unitid)
      for i=1,addon:scanBegin(itemlink) do
        local text = addon:scanText(i)
        if (text:match(ENCHANT_ITEM_MIN_SKILL) or
	    text:match(ENCHANT_ITEM_REQ_LEVEL) or
	    text:match(ENCHANT_ITEM_REQ_SKILL)) and
	    slotid ~= INVSLOT_HAND then -- eng tinkers cant be removed
	  local r,g,b = addon:scanLine(i):GetTextColor()
	  if r > 0.9 and g < 0.15 and b < 0.15 then
            return false, text
	  end
        end
      end
      return true
    end,
  },

  profenchant = {
    order = 60,
    playeronly = true,
    failname = L["Profession Enchant"],
    slots = { "Finger", "Wrist", "Shoulder", "Back" },
    itemcheck = function(itemid, itemlink, slotid)
      local enchant = select(4,addon:decomposeItemLink(itemlink))
      enchant = (enchant and tonumber(enchant)) or 0
      if slotid == INVSLOT_FINGER1 or slotid == INVSLOT_FINGER2 then
        if addon:hasProfession(PID_ENCH) >= 475 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_BACK then
        if addon:hasProfession(PID_TAIL) >= 500 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_WRIST then
        if addon:hasProfession(PID_LW) >= 500 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_SHOULDER then
        if addon:hasProfession(PID_INSC) >= 500 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_HAND then
        if addon:hasProfession(PID_ENG) >= 425 then -- cannot directly check eng chants
          local _,_,hasCD = GetInventoryItemCooldown("player",slotid)
	  return hasCD 
	end
      end
      return true
    end,
  },

  emptysocket = { 
    order = 80,
    failname = L["Empty Socket"],
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
      local empty = addon:countEmptySockets(itemlink)
      return (empty == 0)
    end,
  },

  gemquality = { 
    order = 85,
    failname = L["Gem Quality"],
    desc = L["Require highest quality gems"],
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
       local _,_,_,ilvl = GetItemInfo(itemid)
       local mingemlvl = 85
       local mingemqual = 3
       if ilvl < 285 then
          mingemlvl = 80
          mingemqual = 4
       end
       local s = ""
       for i=1,maxgems do
         local gname,glink = GetItemGem(itemlink, i)
         if glink then 
           local _,_,qual,ilvl = GetItemInfo(glink)
           if ilvl < mingemlvl or qual < mingemqual then 
             if #s > 0 then s = s..", " end
             s = s..glink
           end
         end
       end
       return (#s == 0),s
    end,
  },

  beltbuckle = { 
    order = 90,
    failname = L["Belt Buckle"],
    itemstop = true, -- stop checking this slot on failure
    slots = { "Waist" },
    condition = function(unitid)
      return UnitLevel(unitid) >= 70
    end,
    itemcheck = function(itemid, itemlink, slotid)
      local rawsockets = addon:countRawSockets(itemlink)
      local gems = addon:countGems(itemlink)
      return (gems == rawsockets + 1)
    end,
  },

  meta = { 
    order = 95,
    playeronly = true,
    failname = L["Meta Gem Active"],
    slots = { "Head" },
    itemcheck = function(itemid, itemlink, slotid)
      return not addon:metaInactive(itemlink)
    end,
  },

  profgems = {
    order = 100,
    playeronly = true,
    failname = L["Profession Gemming"],
    condition = function(unitid)
      addon.checktmp.profgems = 0
      return true
    end,
    itemcheck = function(itemid, itemlink, slotid)
      if addon:hasProfession(PID_JC) >= 500 then
        addon.checktmp.profgems = addon.checktmp.profgems + 
	   select(2, addon:countGems(itemlink))
      end
      if addon:hasProfession(PID_BS) >= 400 then
        if slotid == INVSLOT_HAND or slotid == INVSLOT_WRIST then
           local rawsockets = addon:countRawSockets(itemlink)
           local gems = addon:countGems(itemlink)
           return (gems == rawsockets + 1), L["Missing Blacksmith Socket"]
        end
      end
      return true
    end,
    playercheck = function(unitid)
      if addon:hasProfession(PID_JC) >= 500 then
        return (addon.checktmp.profgems == 3), addon.checktmp.profgems.." "..(Lchimera or "")
      end
      return true
    end,
  },

  talentpts = {
    order = 1000,
    failname = L["Unspent Talent Points"],
    checktype = "stat",
    playercheck = function(unitid) 
      local unspent = GetUnspentTalentPoints(not UnitIsUnit(unitid,"player"),false, nil)
      return (not unspent or unspent == 0), unspent and unspent.." "..L["Unspent Talent Points"]
    end,
  },

  glyphs = {
    order = 1010,
    failname = L["Empty Glyph Slots"],
    checktype = "stat",
    playeronly = true,
    playercheck = function(unitid) 
      local empty = 0
      for i = 1, NUM_GLYPH_SLOTS do
        local enabled, _, _, glyphSpell = GetGlyphSocketInfo(i, nil)
        if enabled and not glyphSpell then
	  empty = empty + 1
	end
      end
      if empty > 0 then
	  return false, empty.." "..L["Empty Glyph Slots"]
      else 
          return true
      end
    end,
  },
  
  resilience = {
    order = 1090,
    failname = L["PvP Gear"],
    playeronly = true,
    threshdesc = L["Resilience"],
    checktype = "stat",
    threshmin = 0,
    threshmax = 4000,
    threshstep = 1,
    threshname = L["Maximum PvP resilience to allow"],
    playercheck = function(unitid) 
      local val = GetCombatRating(COMBAT_RATING_RESILIENCE_CRIT_TAKEN)
      local max = settings.thresh.resilience
      return val <= max
    end,
  },

  spellhit =  statCheck(1100, "spellhit", GetSpellMissChance,  L["Spell Hit"], L["Maximum spell miss chance to allow"]),
  meleehit =  statCheck(1110, "meleehit", GetMeleeMissChance,  L["Melee Hit"], L["Maximum melee miss chance to allow"]),
  rangedhit = statCheck(1120, "rangedhit", GetRangedMissChance, L["Ranged Hit"], L["Maximum ranged miss chance to allow"]),
  expertise = statCheck(1130, "expertise", GetEnemyDodgeChance, L["Expertise"], L["Maximum dodge chance to allow"]),

  pet = {
    order = 2000,
    failname = L["Pet Summoned"],
    playeronly = true,
    checktype = "buff",
    aliveonly = true,
    condition = function(unitid)
        if check_class == "HUNTER" or check_class == "WARLOCK" then
	  return true
        elseif check_class == "MAGE" then
	  local tabIndex = GetPrimaryTalentTree(false, false)
          return (tabIndex == 3)
        elseif check_class == "DEATHKNIGHT" then
	  local tabIndex = GetPrimaryTalentTree(false, false)
          return (tabIndex == 3)
	else
	  return false
	end
    end,
    playercheck = function(unitid) 
      if IsMounted() or UnitInVehicle(unitid) then
        return true -- cannot reliably check when mounted
      else
        return UnitExists("playerpet")
      end
    end,
  },

  flask = {
    order = 2005,
    failname = L["Flask"],
    checktype = "buff",
    aliveonly = true,
    default = false,
    condition = function(unitid)
      return UnitLevel(unitid) == MAX_PLAYER_LEVEL_TABLE[#MAX_PLAYER_LEVEL_TABLE]
    end,
    playercheck = function(unitid)
      return hasBuff(unitid, {79469,79470,79471,79472,94160,92725,92729,92730,92731})
    end,
  },

  wellfed = {
    order = 2007,
    failname = L["Well Fed"],
    checktype = "buff",
    aliveonly = true,
    default = false,
    playercheck = function(unitid)
      return hasBuff(unitid, {35272,44106,43730,43722,43763})
    end,
  },

  failbuffs = {
    order = 2010,
    failname = L["Fail Buffs"],
    checktype = "buff",
    aliveonly = true,
    class = { "HUNTER", "PALADIN" },
    playercheck = function(unitid) 
      local crusader = GetSpellInfo(32223)
      local pack = GetSpellInfo(13159)
      local cheetah = GetSpellInfo(5118)
      local caster = select(8, UnitBuff(unitid, crusader))
      if UnitIsUnit(hasBuff(unitid, {32223}) or "none", unitid) then -- crusader
        return false, tostring(GetSpellInfo(32223))
      end
      if UnitIsUnit(hasBuff(unitid, {13159}) or "none", unitid) then -- pack
        return false, tostring(GetSpellInfo(13159))
      end
      if UnitIsUnit(hasBuff(unitid, {5118}) or "none", unitid) then -- cheetah
        return false, tostring(GetSpellInfo(5118))
      end
      return true
    end,
  },

  tankstance = {
    order = 2100,
    failname = L["Tank Stance"],
    checktype = "buff",
    aliveonly = true,
    condition = function(unitid) return (UnitInParty(unitid) and GetNumPartyMembers() > 0) or UnitInRaid(unitid) end,
    playercheck = function(unitid) 
      local istank = addon:raidRole(unitid) == "TANK"
      if hasBuff(unitid, {48263,25780}) then -- Blood presence, RF
        return istank
      elseif check_class == "WARRIOR" and GetShapeshiftFormID() == 18 then
        return istank
      elseif check_class == "DRUID" then -- dont check shapeshift forms
        return true
      else
        return not istank
      end
    end,
  },

  dkheader =     buffHeader(3000, L["Death Knight"]),
  dkhorn       = buffCheck(3005, L["Horn of Winter"], "DEATHKNIGHT", buffids.strength),
  dkpresence   = buffCheck(3007, L["Presence"], "DEATHKNIGHT", {48263, 48266, 48265}),
  druidheader =  buffHeader(3010, L["Druid"]),
  druidmark    = buffCheck(3015, L["Mark of the Wild"], "DRUID", {1126, 20217}), -- GotW, kings
  hunterheader = buffHeader(3020, L["Hunter"]),
  hunteraspect = buffCheck(3025, L["Aspect"], "HUNTER", {13165, 20043, 5118, 13159, 82661}),
  mageheader =   buffHeader(3030, L["Mage"]),
  magearmor =    buffCheck(3035, L["Armor"], "MAGE", {6117, 7302, 30482}),
  mageintel =    buffCheck(3037, L["Arcane Intel"], "MAGE", {1459, 61316}),
  paladinheader =buffHeader(3040, L["Paladin"]),
  paladinseal =  buffCheck(3045, L["Seal"], "PALADIN", {20165, 20164, 31801, 20154}),
  paladinbless = buffCheckTwo(3047, L["Blessing"], "PALADIN", buffids.might, buffids.kings),
  priestheader = buffHeader(3050, L["Priest"]),
  priestarmor =  buffCheck(3052, L["Inner Fire/Will"], "PRIEST", {588, 73413}),
  priestfort   = buffCheck(3055, L["Fortitude"], "PRIEST", {21562}),
  priestresist = buffCheck(3057, L["Shadow Resist"], "PRIEST", {27683, 19726}), -- shadow protect, resist aura
  rogueheader =  buffHeader(3060, L["Rogue"]),
  shamanheader = buffHeader(3070, L["Shaman"]),
  shamanshield = buffCheck(3072, L["Water/Lightning Shield"], "SHAMAN", {324, 52127}),
  warlockheader =buffHeader(3080, L["Warlock"]),
  warlockarmor = buffCheck(3082, L["Armor"], "WARLOCK", {687,  28176}),
  warlocksoullink = buffCheck(3085, L["Soul Link"], "WARLOCK", {19028}),
  warriorheader =buffHeader(3090, L["Warrior"]),
  warriorshout = buffCheckTwo(3095, L["Battle/Commanding Shout"], "WARRIOR", buffids.strength, buffids.fort),

  paladinaura = {
    order =  3048,
    failname = L["Aura"],
    checktype = "buff",
    aliveonly = true,
    playeronly = true,
    class = { "PALADIN" },
    playercheck = function(unitid) 
      return GetShapeshiftForm() ~= 0
    end,
  },

  spriestbuff = {
    order =  3058,
    failname = L["Shadow Buffs"],
    checktype = "buff",
    aliveonly = true,
    playeronly = true,
    condition = function(unitid) return check_class == "PRIEST" and GetPrimaryTalentTree(false, false) == 3 end,
    playercheck = function(unitid) 
      for ti = 1, GetNumTalents(3, false, nil) do
        local link = GetTalentLink(3, ti, false, nil, nil)
        if link:match("\124Htalent::0\124") and not hasBuff(unitid, {15473}) then -- Shadowform
           return false, GetSpellInfo(15473)
        elseif link:match("\124Htalent::0\124") and not hasBuff(unitid, {15286}) then -- Vampiric Embrace
           return false, GetSpellInfo(15286)
        end
      end
      return true
    end,
  },

  roguepoison = {
    order = 3065,
    playeronly = true,
    failname = L["Weapon Poison"],
    threshdesc = L["Minimum minutes remaining on weapon poison to allow"],
    checktype = "buff",
    threshmin = 0,
    threshmax = 15,
    threshstep = 1,
    class = { "ROGUE" },
    playercheck = function(unitid) return checkWeaponImbue(unitid, "roguepoison") end,
  },  
  shamanimbue = {
    order = 3075,
    playeronly = true,
    failname = L["Weapon Imbue"],
    threshdesc = L["Minimum minutes remaining on weapon imbue to allow"],
    checktype = "buff",
    threshmin = 0,
    threshmax = 15,
    threshstep = 1,
    class = { "SHAMAN" },
    playercheck = function(unitid) return checkWeaponImbue(unitid, "shamanimbue") end,
  },
}
end
-- -------------------------------------------------------------------------------------------------
function addon:RunCheck(ismanual)
  debug("RunCheck")
  addon:ResetFails("player")
  addon:CheckUnit("player")
  addon:ReportFails("player", ismanual)
end
local checkcond = {}
function addon:CheckUnit(unitid)
  local isplayer = UnitIsUnit(unitid, "player")
  local _, class = UnitClass(unitid)
  check_class = class
  addon.lowduration = false
  -- preconditions
  wipe(checkcond)
  for _,check in ipairs(addon.checks_ordered) do
     local ctype = check.checktype or "gear"
     local doit = not check.header and settings.checks[check.checkname] and settings.checks[ctype]
     if doit and check.playeronly and not isplayer then
       doit = false
     end
     if doit and check.aliveonly and UnitIsDead(unitid) then
       doit = false
     end
     if doit and check.class then
       if type(check.class) == "string" and check_class ~= check.class then
         doit = false
       elseif type(check.class) == "table" then
         doit = false
         for _,c in pairs(check.class) do
	   if c == check_class then
	     doit = true
	     break
	   end
	 end
       end
     end
     if doit and check.condition then
       checkcond[check] = check.condition(unitid)
     else
       checkcond[check] = doit
     end
  end
  -- item checks
  for name,data in pairs(addon.itemslots) do
    local itemid = GetInventoryItemID(unitid, data.id)
    local itemlink = GetInventoryItemLink(unitid, data.id)

    if not itemid and data.id == INVSLOT_OFFHAND and 
       addon:is2HWeapon(GetInventoryItemID(unitid, INVSLOT_MAINHAND)) --- XXX: titan's grip
    then
      -- offhand is empty for two-hander, skip checks
    else
     for _,check in ipairs(addon.checks_ordered) do
       local doit = checkcond[check]
       if doit and check.slots then
         doit = false
         for _,s in pairs(check.slots) do
	   if s == name or s == data.shortname then
	     doit = true
	     break
	   end
	 end
       end
       if doit and not itemid and check.checkname ~= "empty" then
         doit = false
       end
       if doit and check.itemcheck then
         --debug("Checking "..data.shortname.." "..itemid.." "..itemlink)
         local passed = check.itemcheck(itemid, itemlink, data.id, unitid)
         if not passed then
	   addon:RecordFail(unitid, check, name, itemlink or "")
           if check.itemstop then
	     break
	   end
	 end
       end
     end
    end
  end
  -- player checks
  for _,check in ipairs(addon.checks_ordered) do
     local doit = checkcond[check]
     if doit and check.playercheck then
       local passed, note = check.playercheck(unitid)
       if not note and addon.lowduration then
         note = L["(Low duration)"]
       end
       if not passed then
	 addon:RecordFail(unitid, check, nil, nil, note)
       end
     end
  end
end

addon.failtable = {}
function addon:ResetFails(unitid)
  addon.failtable[unitid] = addon.failtable[unitid] or {}
  wipe(addon.failtable[unitid])
end

function addon:RecordFail(unitid, check, slot, itemlink, note)
  addon.failtable[unitid] = addon.failtable[unitid] or {}
  local faildata = addon.failtable[unitid][check.order] or {}
  faildata.check = check
  faildata.note = note
  if slot then
    faildata.slot = faildata.slot or {}
    faildata.slot[slot] = itemlink
  end
  addon.failtable[unitid][check.order] = faildata
  addon.failtable[unitid].count = (addon.failtable[unitid].count or 0) + 1
end

function addon:ReportFails(unitid, ismanual)
  local ft = addon.failtable[unitid]
  if not ft.count then
      debug("No failures for "..UnitName(unitid))
      if settings.general.alwayschat or ismanual then
        chatAlert(string.format(L["Detected %i failures for %s"],0,UnitName(unitid)))
	chatAlert(L["Congratulations - You're (probably) not a retard!"])
      end
  else
    chatAlert(string.format(L["Detected %i failures for %s"],ft.count,UnitName(unitid))..":")
    for _, check in ipairs(addon.checks_ordered) do
       local faildata = ft[check.order]
       if faildata then
         local str = "\124cffff0000"..faildata.check.failname.."\124r"
         if faildata.slot then
	   local cnt = 0
	   str = str..": "
	   for slot,itemlink in pairs(faildata.slot) do
	     cnt = cnt + 1
	     if cnt > 1 then str = str..", " end
	     str = str..itemlink.." ("..addon.itemslots[slot].shortname..")"
	     if cnt > 2 then chatAlert(str); str = "   "; cnt = 0 end
           end
	   if #str > 10 then
	     chatAlert(str)
	   end
         else
	   if faildata.note then
	     str = str.." "..faildata.note
	   end
           chatAlert(str)
         end
       end
    end
  end
    addon.lastsoundtime = addon.lastsoundtime or 0
    local now = GetTime()
    if settings.general.sound and now - addon.lastsoundtime >= 5 then
      addon.lastsoundtime = now
      local soundfile = addon.soundtable[settings.general.soundname]
      if not ft.count and ismanual then 
        soundfile = addon.soundok
      elseif settings.general.soundrandom then
        local idx = math.random(#addon.soundtable)
	soundfile = addon.soundtable[idx]
      end
      if soundfile then
        PlaySoundFile(soundfile, "Master")
      end
    end
    return ft.count
end
-- -------------------------------------------------------------------------------------------------
local function table_default(t, d)
  for k,v in pairs(d) do
    if type(v) == "table" then
      t[k] = t[k] or {}
      table_default(t[k], v)
    elseif t[k] == nil then
      t[k] = v
    end
  end
end

function addon:RefreshConfig()
  -- things to do after load or settings are reset
  debug("RefreshConfig")
  settings = addon.db.profile
  addon.settings = settings
  for n,c in pairs(addon.checks) do
    if defaults.profile.checks[n] == nil then
      if c.default ~= nil then
        defaults.profile.checks[n] = c.default
      else
        defaults.profile.checks[n] = true
      end
    end
  end
  table_default(settings,defaults.profile)
  for k,v in pairs(defaults.profile) do
     if settings[k] == nil then
       settings[k] = table_clone(v)
     end
  end
  settings.gearchecks = settings.checks
  settings.statchecks = settings.checks
  settings.buffchecks = settings.checks
  settings.loaded = true
  addon:Update()
end

function addon:Update()
  -- things to do when settings change
  if addon.LDBo then
    if settings.minimap.hide then
      minimapIcon:Hide(addonName)
    else
      minimapIcon:Show(addonName)
    end
  end
end

function addon:SetupVersion()
   local svnrev = 0
   local files = addon.svnrev
   files["X-Build"] = tonumber((GetAddOnMetadata(addon.name, "X-Build") or ""):match("%d+"))
   files["X-Revision"] = tonumber((GetAddOnMetadata(addon.name, "X-Revision") or ""):match("%d+"))
   for _,v in pairs(files) do -- determine highest file revision
     if v and v > svnrev then
       svnrev = v
     end
   end
   addon.revision = svnrev

   files["X-Curse-Packaged-Version"] = GetAddOnMetadata(addon.name, "X-Curse-Packaged-Version")
   files["Version"] = GetAddOnMetadata(addon.name, "Version")
   addon.version = files["X-Curse-Packaged-Version"] or files["Version"] or "@"
   if string.find(addon.version, "@") then -- dev copy uses "@.project-version.@"
      addon.version = "r"..svnrev
   end
end


function addon:OnInitialize()
  -- AddonLoader support
  SLASH_RETARDCHECK1 = nil
  SlashCmdList["RETARDCHECK"] = nil
  hash_SlashCmdList["/retardcheck"] = nil
  SLASH_RC1 = nil
  SlashCmdList["RC"] = nil
  hash_SlashCmdList["/rc"] = nil
  GameTooltip:Hide()

  addon:SetupVersion()
  addon.db = LibStub("AceDB-3.0"):New("RetardCheckDB", defaults)
  addon:InitData()
  addon:RefreshConfig()

  local options = addon:myOptions()
  LibStub("AceEvent-3.0"):Embed(addon)
  LibStub("AceConfigRegistry-3.0"):ValidateOptionsTable(options, addonName)
  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, options, {"retardcheck", "rc"})
  optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, addonName, nil, "general")
  optionsFrame.default = function()
       for k,v in pairs(defaults.profile) do settings[k] = table_clone(v) end
       addon:RefreshConfig()
       if InterfaceOptionsFrame:IsShown() then
         addon:Config(); addon:Config()
       end
  end
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Gear Checks"], addonName, "gearchecks").default = optionsFrame.default
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Stat Checks"], addonName, "statchecks").default = optionsFrame.default
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Buff Checks"], addonName, "buffchecks").default = optionsFrame.default
  options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(addon.db)
  profileFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Profiles"], addonName, "profiles")

  debug("OnInitialize")

  self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
  self.db.RegisterCallback(self, "OnDatabaseReset", "RefreshConfig")

end

function addon:Config()
  if optionsFrame then
    if ( InterfaceOptionsFrame:IsShown() ) then
      InterfaceOptionsFrame:Hide();
    else
      InterfaceOptionsFrame_OpenToCategory(profileFrame)
      InterfaceOptionsFrame_OpenToCategory(optionsFrame)
    end
  end
end

function addon:OnEnable()
  debug("OnEnable")
  addon:RegisterEvent("READY_CHECK")
  addon:RegisterEvent("PLAYER_TARGET_CHANGED")
 
  if LDB then
    return
  end
  if AceLibrary and AceLibrary:HasInstance("LibDataBroker-1.1") then
    LDB = AceLibrary("LibDataBroker-1.1")
  elseif LibStub then
    LDB = LibStub:GetLibrary("LibDataBroker-1.1",true)
  end
  if LDB then
    local dataobj = LDB:GetDataObjectByName(addonName) or
     LDB:NewDataObject(addonName, {
        type = "launcher",
        label = addonName,
        icon = "Interface\\AddOns\\"..addonName.."\\icon",
     })
     dataobj.OnClick = function(self, button)
                if button == "RightButton" then
                        addon:Config()
                else
                        addon:RunCheck(true)
                end
        end
     dataobj.OnTooltipShow = function(tooltip)
                if tooltip and tooltip.AddLine then
                        tooltip:SetText(addonName.." "..addon.version)
                        tooltip:AddLine("|cffff8040"..L["Left Click"].."|r "..L["to run a check"])
                        tooltip:AddLine("|cffff8040"..L["Right Click"].."|r "..L["for options"])
                        tooltip:Show()
                end
        end
     addon.LDBo = dataobj
  end 

  if addon.LDBo then
    minimapIcon:Register(addonName, addon.LDBo, settings.minimap)
  end
  addon:Update()
end

-------------------------------------------------------------------------------------
-- Auto checks

addon.lastautocheck = 0
function addon:autocheck() 
  local now = GetTime() 
  if not addon:inBG() and 
     now > addon.lastautocheck+(60*settings.general.autochecktime) then
    addon.lastautocheck = now
    addon:RunCheck()
  end
end

function addon:READY_CHECK(sender)
  if settings.general.readycheck then 
    addon:autocheck()
  end
end

function addon:PLAYER_TARGET_CHANGED()
  local unit = "target"
  if InCombatLockdown() or not UnitExists(unit) then return end
  if settings.general.targetboss then 
    local tgtlvl = UnitLevel(unit)
    local mobtype = UnitClassification(unit)
    if UnitIsEnemy("player",unit) and
       (mobtype == "elite" or mobtype == "rareelite" or mobtype == "worldboss") and
       (tgtlvl == -1 or tgtlvl >= UnitLevel("player")+2) then
      addon:autocheck()
    end
  end
end

